#define SECURITYFILETIMESTAMP 1
#define SECURITYFILERECTANGLE 2

void DrawScene(void);
void DrawTime(void);
void InitHavenGraphics(void);
void ExitVideo(void);
void NextFrame(void);
void InitVideo(void);
void InitGraphics(void);

/*
 * securitylite.c  Copyright (c) 1994 Silicon Graphics Inc.
 * This program may be freely used by any Silicon Graphics customer on
 * Silicon Graphics computers.  It may not be sold or ported to other 
 * vendors hardware without the express written permission of Silicon Graphics.
 *
 * This program demonstrates the ability to combine live video
 * from an IndyCam with traditional IrisGL immediate mode graphics
 * in the same window, controlled by the same process. The program  
 * writes the minimum possible amount of data to the output file to 
 * record what happened within the view of the camera.  The output is
 * black & white and only the parts of the scene that change are 
 * recorded. The resulting file is useful for surveillance but is not
 * terribly 'pretty' since the ambient light changes over time and the 
 * new updated rectangles typically have a slightly different luminance
 * then the one already on the screen.  The program went into operation
 * in the Farmington office on September 28 and we had a break in on 
 * October 1st.  The thief stole over $10000 of new computer gear. On the
 * basis of the video the police were able to get a search warrant and 
 * arrest the guy within 24 hours and we have recovered all of our equipment. 
 * 
 * The data files created by this program are pretty simple and can be read 
 * by the "securityviewer" program.  The key design goals for this program
 * were:
 *
 *  1) Automatic start and stop capability via cron or haven
 *  2) If the program fails at any point the file is readable up to that point.
 *  3) File size is minimized
 *  4) Every frame is timestamped (adds only 4 bytes per frame to the file)
 *
 *
 *
 *
 * -- Barry Brouillette  - 1994
 *
 * Many thanks to those who have coded/archived the base demos for this one
 * 
 *   Eric Miller    - 1994	glvl
 *   Thant Tessman  - 1990	~4Dgifts/examples/grafix/misc/skeleton.c
 *   John Magdziarz - 1993	simpleccapt.c
 *   Wayne Mesard   - 1995      work to make securitylite compatible with haven
 * 
 *  
 *  To run from haven try:
 *     haven /usr/local/bin/securitylite &; sleep 1; xset s activate
 *  This will set up the program to watch your desk when you haven't been 
 *  around for a while.
 */


#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <unistd.h>
#include <gl/gl.h>
#include <vl/vl.h>
#include <sys/utsname.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <gl/gl.h>
#include <gl/device.h>
/* #include <gl/image.h> */

time_t now;
FILE *fp;
int outdataitem;
int LocalScreen = 1;
int lastminx = 0,lastminy = 0,lastmaxx = 0,lastmaxy=0;
int y,x,yy,minx,miny,maxx,maxy;
int swapcount = 0;
int framecount = 0;
int pixelchange=25; /* how much does a pixel have to change before we count it*/
int pixelchangecount = 40;  /* how many pixels change before we record a frame*/
Boolean MACHMODE;
char mydate[100];
char  directoryName[100];
char  fbuf[480][640];   /* buffer which is actually written to data file */
unsigned long  lbuf[308321]; 
unsigned long *lbufp;
short rbuf[640];
short gbuf[640];
short bbuf[640];
short obuf[640];

void CreateFileName(char *name);
void usage(void);
void CalculateJulians(float startdelay,float stopdelay,int *julianstart, int *julianstop);
void GetMyDate(char *mydate);

int diffcount;
char outfilename[100];
int DisplayFrame = 1;
int HavenMode    = 0;
int NapTime      = 0;
int outputToFile = 1;
int forced_node = 0;
int vin = -1;
int julianstart,julianstop;
float startdelay = 0.0, stopdelay = 500.0;
int cliptop    =   0;
int clipbottom = 479;
int clipleft   =   0;
int clipright  = 639;



/* video globals */
    char *_progName;
    VLServer svr;
    VLPath path;
    VLNode src, drn;
    VLControlValue val;
    VLBuffer buffer,buffer0,buffer1,bufferp;
    VLInfoPtr info;
    char *dataPtr;
    char *dataPtr1;
    int c;
    int xsize;
    int ysize;
    long win;

/* Report errors */
void
error_exit(void)
{
    vlPerror(_progName);
    exit(1);
}

void InitVideo()
{

    /* Connect to the daemon */

    if (!(svr = vlOpenVideo("")))
      {
        fprintf(stderr,"vlOpenVideo failed\n");
	error_exit();
      }

    /* Set up a drain node in memory */
    drn = vlGetNode(svr, VL_DRN, VL_MEM, VL_ANY);

    /* Set up a source node on any video source  */
    if(forced_node)
      src = vlGetNode(svr, VL_SRC, VL_VIDEO, vin);
    else
      src = vlGetNode(svr, VL_SRC, VL_VIDEO, VL_ANY);

    /* Create a path using the first device that will support it */
    path = vlCreatePath(svr, VL_ANY, src, drn);

    /* Set up the hardware for and define the usage of the path */
    if ((vlSetupPaths(svr, (VLPathList)&path, 1, VL_SHARE, VL_SHARE)) < 0)
	error_exit();

    /* Set the packing to RGB */
    val.intVal = VL_PACKING_RGB_8; 
    vlSetControl(svr, path, drn, VL_PACKING, &val);

    /* Get the video size */
    vlGetControl(svr, path, drn, VL_SIZE, &val);
    xsize = val.xyVal.x;
    ysize = val.xyVal.y;

    /* Create and register a buffer for 1 frame */
    buffer = vlCreateBuffer(svr, path, drn, 1);
    buffer1 = vlCreateBuffer(svr, path, drn, 1);
    if ((buffer == NULL) || (buffer1 == NULL))
	error_exit();
    vlRegisterBuffer(svr, path, drn, buffer);
    vlRegisterBuffer(svr, path, drn, buffer1);

    /* Begin the data transfer */
    if (vlBeginTransfer(svr, path, 0, NULL))
	error_exit();
}


void NextFrame() {
       ulong *rp,*gp,*bp,*dp;

	do {
	    sginap(1);		/* wait a tick */
	    info = vlGetNextValid(svr, buffer);
	} while (!info);

	/* Get a pointer to the frame */
	dataPtr = vlGetActiveRegion(svr, buffer, info);
	dataPtr1 = vlGetActiveRegion(svr, buffer, info);

        if(NapTime)
          sginap(NapTime);

        dp = (ulong *)dataPtr;
        diffcount = 0;
        minx = xsize;
        maxx = 0;
        miny = ysize;
        maxy = 0;

        for(y=cliptop;y<=clipbottom;y++)
           {
            dp = (ulong *)dataPtr + (xsize * y) + clipleft;
            for(x = clipleft;x<=clipright;x++) /* Unpack the red green and blue values */
             {
              rbuf[x] = *dp & 0x000000FF;
              gbuf[x] = (*dp & 0x0000FF00) >> 8;
              bbuf[x] = (*dp & 0x00FF0000) >> 16;
              dp++;
             }

            /* convert to Black and White for smaller storage requirements */
            rgbrowtobw(rbuf,gbuf,bbuf,obuf,xsize);

            /* Determine how many pixels changed since last frame */
            for(x=clipleft;x<=clipright;x++)
            {
             if((fbuf[y][x] - (char)obuf[x] > pixelchange) || ((char)obuf[x] - fbuf[y][x] > pixelchange)) 
               {
                 diffcount++;
                 if (y < miny)    /* determine box where pixels change */
                    miny = y;
                 if(y > maxy)
                    maxy = y;
                 if(x< minx)
                    minx = x;
                 if(x>maxx)
                    maxx = x;
               }


              fbuf[y][x] = (char)obuf[x];
             }
          }
          if(NapTime)
            sginap(NapTime);

          now = time(0);
          if(now > julianstop)
           {
              if(outputToFile)
                fclose(fp);
              ExitVideo();
              GetMyDate(mydate);
              printf("SecurityLite stopped at %s\n",mydate);
              exit(0);
           }

          if(diffcount > pixelchangecount)
           {


             if(now >= julianstart)
              {/* write whole screen first time through and each 100th frame */
                 if((framecount % 100) == 0)
                   {
                     minx = clipleft;
                     miny = cliptop;
                     maxx = clipright;
                     maxy = clipbottom;
                   }
                 else
                   {               /* expand store area slightly */
                     minx -= 10;
                     miny -= 10;
                     maxx += 10;
                     maxy += 10;
                     if(minx < clipleft) minx = clipleft;
                     if(miny < cliptop) miny = cliptop;
                     if(maxx > (clipright)) maxx = (clipright);
                     if(maxy > (clipbottom)) maxy = (clipbottom);
                   }


             if(DisplayFrame)
               {
                char string[100];
                drawmode(OVERDRAW);
                color(BLACK);
                recti(lastminx,lastminy,lastmaxx,lastmaxy);
                color(RED);
                lastminx = minx;
                lastminy = 500 - (miny+20);
                lastmaxx = maxx;
                lastmaxy = 500 - (maxy+20);
                recti(lastminx,lastminy,lastmaxx,lastmaxy);
                drawmode(NORMALDRAW);
                swapcount = 0;
               }
           
                 if(outputToFile)
                  {
                   outdataitem = SECURITYFILETIMESTAMP;
                   fwrite(&outdataitem,sizeof(int),1,fp);
                   fwrite(&now,sizeof(time_t),1,fp);
                   outdataitem = SECURITYFILERECTANGLE;
                   fwrite(&outdataitem,sizeof(int),1,fp);
                   fwrite(&minx,sizeof(int),1,fp);
                   fwrite(&miny,sizeof(int),1,fp);
                   fwrite(&maxx,sizeof(int),1,fp);
                   fwrite(&maxy,sizeof(int),1,fp);
                   for(y=miny;y<=maxy;y++)
                      fwrite(&fbuf[y][minx],sizeof(char),maxx-minx+1,fp);
                  }
                 framecount++;
               }
           }
	/* Finished with frame, unlock the buffer */
	vlPutFree(svr, buffer);
}

void ExitVideo()
{
    /* End the data transfer */
    vlEndTransfer(svr, path);

    /* Cleanup before exiting */
    vlDeregisterBuffer(svr, path, drn, buffer);
    vlDestroyBuffer(svr, buffer);
    vlDestroyPath(svr, path);
    vlCloseVideo(svr);
}

void InitGraphics()
 { 
                  /* Set up and open a GL window to display the data */


    foreground();
    prefposition(50,xsize+49,50,550);
    win = winopen("SecurityLite");
    RGBmode();
    pixmode(PM_TTOB, 1);
    singlebuffer();
    if (getgdesc(GD_BITS_OVER_SNG_CMODE) < 2)   /* test for overlay planes */
        MACHMODE = PUPDRAW;
    else {
        MACHMODE = OVERDRAW;
        overlay(2);                               /* define/setup overlays */
    }
    drawmode(OVERDRAW);
    ortho2(0,640,0,500);
    mapcolor (0, 0, 0, 0);
    mapcolor (1, 255, 0, 0);
    drawmode(NORMALDRAW);
    gconfig();
    qdevice(ESCKEY);
    qdevice(MIDDLEMOUSE);
    qdevice(LEFTMOUSE);
    qdevice(MOUSEX);
    qdevice(MOUSEY);
    
}


void InitHavenGraphics() 
  {                   
   prefposition(50,xsize+49,50,550);
   win = winopen("HavenSecurityLite");
   RGBmode();
   doublebuffer();
   gconfig();
  }

void DrawTime()
  {
#define MAX_DATEX 400
    static int datex = MAX_DATEX;

    RGBcolor(0,0,0);
    clear();
    ortho2(0,640,0,500);
    mydate[0] = '\0';
    GetMyDate(mydate);
    RGBcolor(0,0,0);
    rectfi(0,480,640,500); 
    RGBcolor(255,255,255);
    cmov2i(datex,485);
    if (--datex == 0) datex = MAX_DATEX;
    charstr(mydate);
    swapbuffers();
  }

void DrawScene()
   {

    if(swapcount ==  5)
     {
       drawmode(OVERDRAW);
       color(BLACK);
       clear();
       drawmode(NORMALDRAW);
     }
    ortho2(0,640,0,500);
    mydate[0] = '\0';
    GetMyDate(mydate);
    RGBcolor(0,0,0);
    rectfi(0,480,640,500); 
    RGBcolor(255,255,255);
    cmov2i(400,485);
    charstr(mydate);

    if((now >= julianstart) && (now <= julianstop))
      {
        RGBcolor(255,255,255);
        cmov2i(10,485);
        charstr("Recording");
      }
    else
      {
        RGBcolor(255,255,255);
        cmov2i(10,485);
        charstr("Not recording");
      }




      if(LocalScreen == 0)
        {
          if(diffcount > pixelchangecount)
           {
           lbufp = lbuf;
           for(y=miny;y<=maxy;y++)
            {
              for(x=minx;x<=maxx;x++)
               {
                 *lbufp++ = fbuf[y][x] + (fbuf[y][x] << 8) + (fbuf[y][x] << 16);
               }
            }
	   lrectwrite(minx,479-maxy,maxx,479-miny,(unsigned long*)lbuf);
           swapcount++;
           }
        }
       else
        {
	 lrectwrite(0,0, xsize-1, ysize-1, (ulong *)dataPtr);
         swapcount++;
        }
}


int main(int argc, char **argv) {
    long dev;
    short val;

    GetMyDate(mydate);
    printf("SecurityLite started at %s\n",mydate);
    while ((c = getopt(argc, argv, "a:b:c:d:f:hkl:no:p:r:Ss:t:v:z")) != EOF) 
    {
	switch (c) 
	{
            case 'a': 
                startdelay = atof(optarg);
                printf("Program will start recording after %2.1f hours\n",startdelay);
                break;
	    case 'b':
		clipbottom = atoi(optarg);
                printf("Bottom of frame is at %d (479 is very bottom)\n",clipbottom);
                break;
	    case 'c':
		pixelchangecount = atoi(optarg);
                printf("pixelchangecount threshhold is %d '%s'\n",pixelchangecount,optarg);
                break;
            case 'd':
                strcpy(directoryName,optarg);
                printf("File will be put in the '%s' directory\n",directoryName);
                break;
	    case 'f':
		sprintf(outfilename,"%s",optarg);
	        break;
	    case 'k':
		outputToFile = 0;
                printf("No file will be written\n");
	        break;
	    case 'l':
		clipleft = atoi(optarg);
                printf("Left of frame is at %d\n",clipleft);
                break;
	    case 'n':
		DisplayFrame = 0;
                printf("No graphical output\n");
                break;
            case 'o': 
                stopdelay = atof(optarg);
                printf("Program will stop recording after %2.1f hours\n",stopdelay);
                break;
	    case 'p':
		pixelchange = atoi(optarg);
                printf("pixelchange threshhold is %d '%s'\n",pixelchange,optarg);
                break;
	    case 'r':
		clipright = atoi(optarg);
                printf("Right of frame is at %d\n",clipright);
                break;
            case 'S':
                DisplayFrame = 0;
                HavenMode    = 1;
                break;
            case 's': 
                NapTime = atof(optarg);
                printf("Program will delay %d ticks between frames\n",NapTime);
                break;
	    case 't':
		cliptop = atoi(optarg);
                printf("Top of frame is at %d (0 is very top)\n",cliptop);
                break;
	    case 'v':
		forced_node = 1;
		vin = atoi(optarg);
                printf("Video device %d will be used\n",vin);
	        break;
            case 'z':
                LocalScreen = 0;
                printf("Only changed part of scene will be updated on screen.\n");
	        break;
	    default:
	    case 'h':
		usage();
		exit(0);
	        break;
	}
    }

    InitVideo();
    CalculateJulians(startdelay,stopdelay,&julianstart,&julianstop);
    if(DisplayFrame)
       InitGraphics();            /* set up graphics and initialize variables */

    if(HavenMode)
       InitHavenGraphics();


    if(strlen(outfilename) == 0)
       CreateFileName(outfilename);

    if(outputToFile)
     {
      fp = fopen(outfilename,"wb");
      if(fp == NULL)
        {
          printf("Could not open file '%s'\n",outfilename);
          exit(1);
        }
     }

    while (TRUE) {                            /* main input processing loop */

        /*
         * First, if we need to redraw the scene for any reason, redraw
         * it.  
         */
        NextFrame();      /* collect the video */

        if(HavenMode)
          DrawTime();

        if (DisplayFrame)
          {
            DrawScene();

              while (qtest()) {

               switch(dev=qread(&val)) {

                case ESCKEY:
                    if (val)                        /* exit on up, not down */
                        break;
                    exit(0);

                case REDRAW:
                    reshapeviewport();
                    break;

            }    /* End of switch */
           }
        }   
    }                    /* End of main (infinite) loop */
}




void usage(void)
{
printf("USAGE: securitylite [-v videonode] [-f filename] [-o hours till stop] ... other options below\n");
printf("\nOption flags:\n");
printf("-a hours    (number of hours to wait before recording starts)\n");
printf("-o hours    (number of hours to wait before recording stops)\n");
printf("-n          ( Do not open graphical window )\n");
printf("-k          ( Kill file output. No data file will be created )\n");
printf("-f filename (Filename of output file. default name built from system name and date)\n");
printf("-p <value>  ( How much change before pixel is counted as changed - default is 25)\n");
printf("-c <value>  ( How many pixels must change before we write frame - default is 40)\n");
printf("-d <directory>  ( What directory should the file be put in as in -d \"/security/\" )\n");
printf("-v <device> ( Which video device. On an Indy use 0 for the IndyCam and 1 for external video cam )\n");
printf("-b <value>  ( Bottom clip value (479 is the very bottom of the screen) )\n");
printf("-t <value>  ( Top clip value (0 is the very top of the screen) )\n");
printf("-l <value>  ( Left clip value (0 is all the way to the left of the screen) )\n");
printf("-r <value>  ( Right clip value (639 is all the way to the right of the screen) )\n");
printf("-S          ( Screen save mode (for use with haven))\n");
printf("-s <value>  ( Sleep time. Use 30 if you're using the system. Experiment with value to find best mix of           frame rate versus disk usage. )\n");
printf("-z          (When displaying image only show changed part - this is faster if   you are viewing image over a network)\n");
}


void GetMyDate(char *mydate)
{
char *cp;
now = time(0);
cp = ctime(&now);
strcpy(mydate,cp);
return;
}



void CalculateJulians(float startdelay,float stopdelay,int *julianstart,int *julianstop)
{
  time_t now;

  now = time(0);
  *julianstart = now+(time_t)(startdelay * 3600.0);
  *julianstop  = now+(time_t)(stopdelay  * 3600.0);
} 


void CreateFileName(char *name)
{
struct utsname sysname;
char *cp;
time_t now;
char temp[100];

uname(&sysname);
now = time(0);
cp = ctime(&now);
strcpy(temp,cp);

if(strlen(directoryName))
  strcpy(name,directoryName);
else
  name[0] = '\0';

if(temp[8] == ' ')
  temp[8] = '0';
strcat(name,sysname.nodename);
strcat(name,"_");
strncat(name,&temp[4],3);
strcat(name,"_");
strncat(name,&temp[8],2);
strcat(name,"_");
strncat(name,&temp[11],2);
strcat(name,"_");
strncat(name,&temp[14],2);
}

